//! A logger abstraction for the CI broker.

use std::{path::Path, process::ExitStatus, time::Duration};

use clap::ValueEnum;
use radicle::{git::raw::Oid, identity::RepoId, node::Event};
use tracing::{debug, error, info, trace, warn, Level};
use tracing_subscriber::{fmt, layer::SubscriberExt, util::SubscriberInitExt, EnvFilter};

use crate::{
    ci_event::CiEvent,
    ci_event_source::CiEventSource,
    config::Config,
    db::{QueueId, QueuedCiEvent},
    msg::Request,
    node_event_source::NodeEventSource,
    run::Run,
};

// A unique type or kind for each log message.
enum Kind {
    Debug,
    Startup,
    Shutdown,
    GotEvent,
    StartRun,
    AdapterMessage,
    FinishRun,
}

impl std::fmt::Display for Kind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let s = match self {
            Self::Debug => "debug",
            Self::Startup => "startup",
            Self::Shutdown => "shutdown",
            Self::GotEvent => "got_event",
            Self::StartRun => "start_run",
            Self::FinishRun => "finish_run",
            Self::AdapterMessage => "adapter_message",
        };

        write!(f, "{s}")
    }
}

// We define our own type for log levels so that we can apply
// clap::ValueEnum on it.
#[derive(ValueEnum, Eq, PartialEq, Copy, Clone, Debug)]
pub enum LogLevel {
    Trace,
    Debug,
    Info,
    Warning,
    Error,
}

impl std::fmt::Display for LogLevel {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let s = match self {
            Self::Trace => "trace",
            Self::Debug => "debug",
            Self::Info => "info",
            Self::Warning => "warn",
            Self::Error => "error",
        };

        write!(f, "{s}")
    }
}

impl From<LogLevel> for tracing::Level {
    fn from(log_level: LogLevel) -> Self {
        match log_level {
            LogLevel::Trace => Level::TRACE,
            LogLevel::Debug => Level::DEBUG,
            LogLevel::Info => Level::INFO,
            LogLevel::Warning => Level::WARN,
            LogLevel::Error => Level::ERROR,
        }
    }
}

#[allow(clippy::unwrap_used)]
pub fn open(level: LogLevel) {
    #[cfg(test)]
    let writer = fmt::TestWriter::new();

    #[cfg(not(test))]
    let writer = std::io::stderr;

    let fmt_layer = fmt::layer().with_target(false).with_writer(writer).json();

    let filter_layer = EnvFilter::try_new(level.to_string()).unwrap();

    tracing_subscriber::registry()
        .with(filter_layer)
        .with(fmt_layer)
        .init();
}

pub fn start_cib() {
    info!(
        kind = %Kind::Startup,
        version = env!("GIT_HEAD"),
        "CI broker starts"
    );
}

pub fn end_cib_successfully() {
    info!(
        kind = %Kind::Shutdown,
        success = true,
        "CI broker ends successfully"
    );
}

pub fn end_cib_in_error() {
    error!(
        kind = %Kind::Shutdown,
        success = false,
        "CI broker ends in unrecoverable error"
    );
}

pub fn node_event_source_created(source: &NodeEventSource) {
    debug!(
        kind = %Kind::Startup,
        node_event_source = ?source,
        "created node event source"
    );
}

pub fn node_event_source_got_event(event: &Event) {
    trace!(
        kind = %Kind::GotEvent,
        node_event = ?event,
        "node event source received event"
    );
}

pub fn node_event_source_eof(source: &NodeEventSource) {
    debug!(
        kind = %Kind::Shutdown,
        node_event_source = ?source,
        "node event source end of file on control socket"
    );
}

pub fn ci_event_source_created(source: &CiEventSource) {
    debug!(kind = %Kind::Startup, ?source, "created CI event source");
}

pub fn ci_event_source_got_events(events: &[CiEvent]) {
    trace!(
        kind = %Kind::GotEvent,
        ?events,
        "CI event source received events"
    );
}

pub fn ci_event_source_disconnected() {
    info!(kind = %Kind::Shutdown, "CI event source received disconnection");
}

pub fn ci_event_source_end() {
    info!(
        kind = %Kind::Debug,
        "CI event source was notified end of events"
    );
}

pub fn ci_event_source_eof(source: &CiEventSource) {
    info!(kind = %Kind::Debug, ?source, "CI event source end of file");
}

pub fn loaded_config(config: &Config) {
    debug!(kind = %Kind::Startup, ?config, "loaded configuration");
}

pub fn adapter_config(config: &Config) {
    trace!(kind = %Kind::Debug, ?config, "adapter configuration");
}

pub fn queueproc_start() {
    debug!(kind = %Kind::Startup, "start thread to process events until a shutdown event");
}

pub fn queueproc_end() {
    debug!(kind = %Kind::Debug, "thread to process events ends");
}

pub fn queueproc_channel_disconnect() {
    info!(kind = %Kind::Debug, "event notification channel disconnected");
}

pub fn queueproc_queue_length(len: usize) {
    trace!(kind = %Kind::Debug, ?len, "event queue length");
}

pub fn queueproc_picked_event(id: &QueueId, event: &QueuedCiEvent) {
    info!(kind = %Kind::GotEvent, ?id, ?event, "picked event from queue");
}

pub fn queueproc_remove_event(id: &QueueId) {
    info!(kind = %Kind::Debug, ?id, "remove event from queue");
}

pub fn queueproc_action_run(rid: &RepoId, oid: &Oid) {
    info!(kind = %Kind::Debug, ?rid, ?oid, "Action: run");
}

pub fn queueproc_action_shutdown() {
    info!(kind = %Kind::Debug, "Action: shutdown");
}

pub fn queueadd_start() {
    debug!(kind = %Kind::Debug, "start thread to add events from node to event queue");
}

pub fn queueadd_control_socket_close() {
    info!(kind = %Kind::Debug, "no more events from node control socket");
}

pub fn queueadd_push_event(event: &CiEvent) {
    debug!(kind = %Kind::GotEvent, ?event, "insert broker event into queue");
}

pub fn queueadd_end() {
    debug!(kind = %Kind::Debug, "thread to process events ends");
}

pub fn pages_directory_unset() {
    warn!(kind = %Kind::Debug, "not writing HTML report pages as output directory has not been set");
}

pub fn pages_interval(interval: Duration) {
    trace!(
        kind = %Kind::Debug,
        interval = interval.as_secs(),
        "interval for waiting between HTML page updates",
    );
}

pub fn pages_disconnected() {
    info!(kind = %Kind::Debug, "page updater: run notification channel disconnected");
}

pub fn pages_start() {
    debug!(kind = %Kind::Debug, "start page updater thread");
}

pub fn pages_end() {
    debug!(kind = %Kind::Debug, "end page updater thread");
}

pub fn event_disconnected() {
    info!(kind = %Kind::Debug, "connection to node control socket broke");
}

pub fn event_end() {
    info!(kind = %Kind::Debug, "no more node events from control socket: iterator ended");
}

pub fn broker_db(filename: &Path) {
    info!(kind = %Kind::Startup, filename = %filename.display(), "broker database");
}

pub fn broker_start_run(trigger: &Request) {
    info!(kind = %Kind::StartRun, "start CI run");
    debug!(kind = %Kind::Debug, event = ?trigger, "trigger event");
}

pub fn broker_end_run(run: &Run) {
    info!(kind = %Kind::FinishRun, "Finish CI run");
    debug!(kind = %Kind::Debug, ?run, "finished CI run");
}

pub fn adapter_no_first_response() {
    error!(kind = %Kind::AdapterMessage, "no first response message");
}

pub fn adapter_no_second_response() {
    error!(kind = %Kind::AdapterMessage, "no second response message");
}

pub fn adapter_too_many_responses() {
    error!(kind = %Kind::AdapterMessage, "too many response messages");
}

pub fn adapter_stderr_line(line: &str) {
    debug!(kind = %Kind::Debug, stderr_line = line, "adapter stderr");
}

pub fn adapter_result(exit: i32) {
    debug!(kind = %Kind::Debug, exit_code = exit, "adapter exit code");
}

pub fn adapter_did_not_exit_voluntarily() {
    warn!(kind = %Kind::Debug, "adapter did not exit voluntarily: terminated for taking too long");
}

pub fn adapter_did_not_exit() {
    warn!(kind = %Kind::Debug, "adapter did not exit: probably killed by signal");
}

pub fn timeoutcmd_request_termination(result: Result<(), std::sync::mpsc::SendError<()>>) {
    trace!(kind = %Kind::Debug, ?result, "request termination of child process");
}

pub fn timeoutcmd_wait_word_from_nanny() {
    trace!(kind = %Kind::Debug, "wait: wait for word from nanny");
}

pub fn timeoutcmd_wait_got_word_from_nanny() {
    trace!(kind = %Kind::Debug, "got word from nanny");
}

pub fn timeoutcmd_wait_on_nanny_to_end() {
    trace!(kind = %Kind::Debug, "wait: wait on nanny thread to end");
}

pub fn timeoutcmd_wait_on_stdin_writer_to_end() {
    trace!(kind = %Kind::Debug, "wait: wait for stdin writer to terminate");
}

pub fn timeoutcmd_wait_on_stdout_reader_to_end() {
    trace!(kind = %Kind::Debug, "wait: wait for stdout reader to terminate");
}

pub fn timeoutcmd_wait_on_stderr_reader_to_end() {
    trace!(kind = %Kind::Debug, "wait: wait for stderr reader to terminate");
}

pub fn timeoutcmd_wait_on_child_to_end() {
    trace!(kind = %Kind::Debug, "wait: wait for child to terminate");
}

pub fn timeoutcmd_wait_status(status: ExitStatus) {
    trace!(kind = %Kind::Debug, ?status, "wait: wait status");
}

pub fn timeoutcmd_ok() {
    trace!(kind = %Kind::Debug, "wait: return Ok result");
}

pub fn timeoutcmd_nanny_start() {
    trace!(kind = %Kind::Debug, "nanny: start monitoring child");
}

pub fn timeoutcmd_nanny_terminated_as_requested(result: Result<(), std::io::Error>) {
    trace!(kind = %Kind::Debug, ?result, "nanny: terminated child by request");
}

pub fn timeoutcmd_nanny_too_long(
    id: u32,
    elapsed: Duration,
    max: Duration,
    result: Result<(), std::io::Error>,
) {
    trace!(
        kind = %Kind::Debug,
        %id,
        elapsed_ms = elapsed.as_millis(),
        max_ms = max.as_millis(),
        ?result,
        "nanny: child has run for too long",
    );
}

pub fn timeoutcmd_nanny_child_died() {
    trace!(kind = %Kind::Debug, "nanny: child has terminated");
}

pub fn timeoutcmd_nanny_time_to_end() {
    trace!(kind = %Kind::Debug, "nanny: tell other threads it's time to end");
}

pub fn timeoutcmd_nanny_ends() {
    trace!(kind = %Kind::Debug, "nanny: ends");
}

pub fn timeoutcmd_line_reader_try_byte(thread: &'static str) {
    trace!(kind = %Kind::Debug, %thread, "line receiver: try to receive next byte");
}

pub fn timeoutcmd_line_reader_tried_byte(
    thread: &'static str,
    result: Result<u8, std::sync::mpsc::TryRecvError>,
) {
    trace!(kind = %Kind::Debug, ?result, %thread, "line receiver: tried to read line");
}

pub fn timeoutcmd_line_reader_got_line(thread: &'static str, line: &str) {
    trace!(kind = %Kind::Debug, ?line, %thread, "line-receiver: received line");
}

pub fn timeoutcmd_line_reader_got_disconnected(thread: &'static str) {
    trace!(
        kind = %Kind::Debug,
        %thread,
        "line-receiver: disconnected");
}

pub fn timeoutcmd_line_reader_did_child_die(thread: &'static str) {
    trace!(%thread, "line-receiver: has child terminated?");
}

pub fn timeoutcmd_line_reader_child_died(thread: &'static str) {
    trace!(
        kind = %Kind::Debug,
        %thread,
        "line receiver: OK: child has terminated, not returning line",
    );
}

pub fn timeoutcmd_line_reader_child_channel_disconnected(thread: &'static str) {
    trace!(
        kind = %Kind::Debug,
        %thread,
        "line receiver: Disconnected: child has terminated, not returning line",
    );
}

pub fn timeoutcmd_nonblocking_try_byte(thread: &'static str, count: usize) {
    trace!(
        kind = %Kind::Debug,
        received_so_far = count,
        %thread,
        "read_to_end: try to receive next byte",
    );
}

pub fn timeoutcmd_nonblocking_tried_byte(
    thread: &'static str,
    result: &Result<usize, std::io::Error>,
    byte: &[u8],
) {
    trace!(
        kind = %Kind::Debug,
        ?result,
        ?byte,
        %thread,
        "read_to_end: tried to receive byte",
    );
}

pub fn timeoutcmd_nonblocking_eof(thread: &'static str) {
    trace!(kind = %Kind::Debug, %thread, "read_to_end: got end of file");
}

pub fn timeoutcmd_nonblocking_got_too_much(
    thread: &'static str,
    result: Result<usize, std::io::Error>,
    byte: &[u8],
) {
    trace!(
        kind = %Kind::Debug,
        ?result,
        ?byte,
        %thread,
        "read_to_end: received too much",
    );
}

pub fn timeoutcmd_nonblocking_read_error(thread: &'static str, err: &std::io::Error) {
    trace!(kind = %Kind::Debug, ?err, %thread, "read_to_end: read error");
}

pub fn timeoutcmd_nonblocking_ends(thread: &'static str) {
    trace!(kind = %Kind::Debug, %thread, "read_to_end: ends");
}

pub fn debug(msg: &str) {
    debug!(kind = %Kind::Debug, "{msg}");
}

pub fn debug2(msg: String) {
    debug!(kind = %Kind::Debug, "{msg}");
}

pub fn trace(msg: &str) {
    trace!(kind = %Kind::Debug, "{msg}");
}

pub fn trace2(msg: String) {
    trace!(kind = %Kind::Debug, "{msg}");
}

pub fn error(msg: &str, e: &impl std::error::Error) {
    error!("{msg}: {e}");
    let mut e = e.source();
    while let Some(source) = e {
        error!("caused by: {}", source);
        e = source.source();
    }
}
